import { Message, MessageCode } from "../util/message";

import { BaseParam } from "./param-base";
import { ParamDomain, ParamDomainValue } from "./param-domain";
import { ParamValueMode } from "./param-value-mode";
import { Result } from "../util/result";

/**
 * calculabilité du paramètre
 */
export enum ParamCalculability {
    /**
     * paramètre fixé (immuable, cad non modifiable après création)
     */
    NONE,

    /**
     * paramètre libre (modifiable)
     */
    FREE,

    /**
     * paramètre calculable analytiquement, par méthode de Newton, ...
     */
    EQUATION,

    /**
     * paramètre calculable par dichotomie
     */
    DICHO
}

/**
 * définition d'un paramètre d'un noeud de calcul
 */
// tslint:disable-next-line:max-classes-per-file
export class ParamDefinition extends BaseParam {
    /**
     * calculabilité
     */
    private _calc: ParamCalculability;

    constructor(s: string, d: ParamDomain | ParamDomainValue, val?: number) {
        super(s, d, val);
        this._calc = ParamCalculability.FREE;
    }

    get v(): number {
        if (this.isReferenceDefined) {
            const ro = this.referencedObject;

            if (ro instanceof BaseParam) {
                switch (this.referencedParamValues.valueMode) {
                    case ParamValueMode.CALCUL:
                        const r = this.referencedResult;
                        if (r.nbResultElements == 1)
                            return r.resultElement.vCalc;
                        throw new Error(`il n'y a pas exactement un ResultElement dans le Result "${r.name}"`);

                    default:
                        return this.referencedParamValues.currentValue;
                }
            }
            else if (ro instanceof Result)
                return ro.vCalc;

            return this.referencedExtraResult;
        }

        return super.getValue();
    }

    set v(val: number) {
        if (this.calculability === ParamCalculability.NONE) {
            const e = new Message(MessageCode.ERROR_PARAMDEF_VALUE_FIXED);
            e.extraVar.symbol = this.symbol;
            throw e;
        }
        super.setValue(val);
    }

    /*
     * méthodes de calculabilité
     */

    /**
     * variable calculable par l'équation ?
     */
    public isAnalytical(): boolean {
        return this.calculability === ParamCalculability.EQUATION;
    }

    get calculability(): ParamCalculability {
        if (this._calc === undefined) {
            //   throw "value of parameter '" + this._symbol + "' calculability is not defined";
            const e = new Message(MessageCode.ERROR_PARAMDEF_CALC_UNDEFINED);
            e.extraVar.symbol = this.symbol;
            throw e;
        }

        return this._calc;
    }

    set calculability(c: ParamCalculability) {
        this._calc = c;
    }

    public clone(): ParamDefinition {
        const res = new ParamDefinition(this.symbol, this.getDomain().clone(), this.uncheckedValue);
        res._calc = this._calc;
        return res;
    }
}
