import { Pair } from "../util/pair"
import { DefinedNumber } from "../util/definedvalue";
import { IReferencedNub, INubReference, NubReference } from "../value_ref/object_ref";
import { Result } from "..";
import { ParamValueMode } from "./param-value-mode";
import { ParamValueIterator, IterableValues, NamedIterableValues, NumberIterator } from "./param-value-iterator";

export class ParamValues implements INubReference, IterableValues {
    /**
     * mode de génération des valeurs : min/max, liste, ...
     */
    private _valueMode: ParamValueMode;

    /**
     * valeur numérique (éventuellement non définie) dans le mode "valeur unique" (ParamValueMode.SINGLE)
     */
    private _singleValue: DefinedNumber;

    /**
     * valeur min dans le cas ParamValueMode.MINMAX
     */
    private _minValue: number = undefined;

    /**
     * valeur max dans le cas ParamValueMode.MINMAX
     */
    private _maxValue: number = undefined;

    /**
     * pas de progression dans le cas ParamValueMode.MINMAX
     */
    private _stepValue: number = undefined;

    /**
     * liste de valeurs dans le cas ParamValueMode.LISTE
     */
    private _valueList: number[];

    /**
     * valeur courante (éventuellement non définie) indépendemment du mode
     */
    private _currentValue: DefinedNumber;

    /**
     * itérateur courant
     */
    private _iterator: NumberIterator;

    /**
     * implémentation par délégation de INubReference
     */
    private _nubRef: NubReference;

    constructor() {
        this._singleValue = new DefinedNumber();
        this._currentValue = new DefinedNumber();
        this._nubRef = new NubReference();
    }

    public setValues(o: number | any, max?: number, step?: number) {
        if (typeof (o) === "number") {
            if (max == undefined) {
                this._valueMode = ParamValueMode.SINGLE;
                this._singleValue.value = o as number;
                this._currentValue.value = o as number;
            } else {
                this._valueMode = ParamValueMode.MINMAX;
                this._minValue = o as number;
                this._maxValue = max;
                this._stepValue = step;
                this._currentValue.undefine();
            }
        }
        else if (Array.isArray(o)) {
            this._valueMode = ParamValueMode.LISTE;
            this._valueList = o;
            this._currentValue.undefine();
        }
        else
            throw new Error(`ParamValues.setValues() :  appel invalide`);
    }

    public get valueMode() {
        return this._valueMode;
    }

    public set valueMode(m: ParamValueMode) {
        this._valueMode = m;
    }

    private checkValueMode(expected: ParamValueMode) {
        if (this._valueMode != expected)
            throw new Error(`ParamValues : mode de valeurs ${ParamValueMode[expected]} incorrect`);
    }

    public check() {
        switch (this._valueMode) {
            case ParamValueMode.SINGLE:
                if (!this._singleValue.isDefined)
                    throw new Error(`ParamValues : valeur fixe non définie`);
                break;

            case ParamValueMode.MINMAX:
                if (this._minValue == undefined)
                    throw new Error(`ParamValues : valeur min non définie`);
                if (this._maxValue == undefined)
                    throw new Error(`ParamValues : valeur max non définie`);
                if (this._stepValue == undefined)
                    throw new Error(`ParamValues : valeur du pas non définie`);
                if (this._minValue > this._maxValue)
                    throw new Error(`ParamValues : min > max`);
                break;

            case ParamValueMode.LISTE:
                if (this._valueList == undefined)
                    throw new Error(`ParamValues : liste de valeurs non définie`);
                break;

            case ParamValueMode.LINK:
            case ParamValueMode.CALCUL:
                break;

            default:
                throw new Error(`ParamValues : mode de valeurs ${ParamValueMode[this._valueMode]} incorrect`);
        }
    }

    /**
     * valeur courante
     */
    public get currentValue(): number {
        if (this.isReferenceDefined)
            return this._nubRef.referencedParamValues.currentValue;
        return this._currentValue.value;
    }

    public set currentValue(v: number) {
        this._currentValue.value = v;
    }

    /**
     * valeur dans le mode SINGLE
     */
    public get singleValue(): number {
        if (this.isReferenceDefined)
            return this._nubRef.referencedParamValues.singleValue;
        return this._singleValue.value;
    }

    public get uncheckedValue() {
        return this._currentValue.uncheckedValue;
    }

    public setSingleValue(v: number) {
        this._valueMode = ParamValueMode.SINGLE;
        this._singleValue.value = v;
        this._currentValue.value = v;
    }

    public get isDefined() {
        return this._currentValue.isDefined;
    }

    public get min() {
        switch (this._valueMode) {
            case ParamValueMode.MINMAX:
                return this._minValue;

            case ParamValueMode.LINK:
                return this._nubRef.referencedParamValues.min;

            default:
                this.checkValueMode(ParamValueMode.MINMAX); // pour générer une erreur
                return undefined; // pour le compilo
        }
    }

    public set min(v: number) {
        this._minValue = v;
        this._valueMode = ParamValueMode.MINMAX;
    }

    public get max() {
        switch (this._valueMode) {
            case ParamValueMode.MINMAX:
                return this._maxValue;

            case ParamValueMode.LINK:
                return this._nubRef.referencedParamValues.max;

            default:
                this.checkValueMode(ParamValueMode.MINMAX);
                return undefined; // pour le compilo
        }
    }

    public set max(v: number) {
        this._maxValue = v;
        this._valueMode = ParamValueMode.MINMAX;
    }

    public get stepRefValue(): Pair {
        this.checkValueMode(ParamValueMode.MINMAX);
        return new Pair(1e-9, this._maxValue - this._minValue);
    }

    public get step() {
        switch (this._valueMode) {
            case ParamValueMode.MINMAX:
                return this._stepValue;

            case ParamValueMode.LINK:
                return this._nubRef.referencedParamValues.step;

            default:
                this.checkValueMode(ParamValueMode.MINMAX); // pour générer une erreur
                return undefined; // pour le compilo
        }
    }

    public set step(v: number) {
        this._stepValue = v;
        this._valueMode = ParamValueMode.MINMAX;
    }

    public get valueList() {
        this.checkValueMode(ParamValueMode.LISTE);
        return this._valueList;
    }

    public set valueList(l: number[]) {
        this._valueList = l;
        this._valueMode = ParamValueMode.LISTE;
    }

    /**
     * crée un ParamValueIterator
     * @param reverse true si on veut itérer max->min ou depuis la fin de la liste
     */
    public getValuesIterator(reverse: boolean = false): NumberIterator {
        return new ParamValueIterator(this, reverse);
    }

    /**
     * 
     * @param reverse prépare un itérateur pour parcourir les valeurs
     */
    public initIterator(reverse: boolean = false) {
        switch (this._valueMode) {
            case ParamValueMode.LISTE:
            case ParamValueMode.MINMAX:
                this._iterator = this.getValuesIterator(reverse);
                break;

            case ParamValueMode.LINK:
                this._iterator = this._nubRef.referencedParamValues.getValuesIterator(reverse);
                break;

            default:
                throw new Error(`ParamValues : mode de valeurs ${ParamValueMode[this._valueMode]} incorrect`);
        }
    }

    /**
     * @return true si il reste des valeurs à parcourir par l'itérateur courant
     */
    public get hasNext(): boolean {
        return this._iterator.hasNext;
    }

    /**
     * fixe la valeur courante à la prochaine valeur à parcourir par l'itérateur courant
     * @return prochaine valeur à parcourir par l'itérateur courant
     */
    public get next(): number {
        this._currentValue.value = this._iterator.next().value;
        return this._currentValue.value;
    }

    // interface INubReference

    public defineReference(target: IReferencedNub, desc: string) {
        this.valueMode = ParamValueMode.LINK;
        this._nubRef.defineReference(target, desc);
    }

    public undefineReference() {
        this._nubRef.undefineReference();
    }

    public get referencedNub(): IReferencedNub {
        return this._nubRef.referencedNub;
    }

    public get referenceDefinition(): string {
        return this._nubRef.referenceDefinition;
    }

    public get isReferenceDefined(): boolean {
        return this._valueMode === ParamValueMode.LINK && this._nubRef.isReferenceDefined;
    }

    public get referencedParamValues(): ParamValues {
        return this._nubRef.referencedParamValues;
    }

    public get referencedResult(): Result {
        return this._nubRef.referencedResult;
    }

    public get referencedExtraResult(): any {
        return this._nubRef.referencedExtraResult;
    }

    public get referencedValuesIterator(): any {
        return this._nubRef.referencedValuesIterator;
    }

    public get referencedObject(): NamedIterableValues {
        if (this.isReferenceDefined)
            return this._nubRef.referencedObject;
        return undefined;
    }

    // interface IterableValues

    public get valuesIterator(): NumberIterator {
        return this.getValuesIterator();
    }

    public get hasMultipleValues(): boolean {
        if (this.isReferenceDefined)
            var it = this.referencedValuesIterator;
        else {
            // dans certains cas (mode LINK mais aucune valeur liable compatible), on ne peut avoir d'itérateur
            if (this._valueMode !== ParamValueMode.LINK)
                it = this.getValuesIterator();
        }

        if (it === undefined)
            return false;

        let n = 0;
        for (const v of it) {
            n++;
            if (n > 1)
                break;
        }
        return n > 1;
    }
}
