import { RectangularStructureParams } from "./rectangular_structure_params";
import { Structure } from "./structure";
import { StructureCem88d } from "./structure_cem88d";
import { StructureCem88v } from "./structure_cem88v";
import { StructureCunge80 } from "./structure_cunge80";
import { StructureKivi } from "./structure_kivi";
import { StructureKiviParams } from "./structure_kivi_params";
import { StructureOrificeFree } from "./structure_orifice_free";
import { StructureOrificeSubmerged } from "./structure_orifice_submerged";
import { StructureWeirFree } from "./structure_weir_free";
import { StructureType, LoiDebit, StructureProperties } from "./structure_props";

export function CreateStructure(structureType: StructureType, loiDebit: LoiDebit, dbg: boolean = false): Structure {
    const rectStructPrms: RectangularStructureParams = new RectangularStructureParams(
        0,  // Q
        100,        // ZDV
        102,        // Z1
        101.5,      // Z2
        2,          // L
        0.4        // Cd pour un seuil rectangulaire
        // W = Infinity par défaut pour un seuil
    );
    switch (structureType) {
        case StructureType.VanneRectangulaire:
            rectStructPrms.W.v = 0.5;
            rectStructPrms.Cd.v = 0.6; // Cd pour une vanne rectangulaire
            if (!(StructureProperties.isCompatibleValues(StructureType.VanneRectangulaire, loiDebit))) {
                throw new Error(
                    `la loi de débit ${LoiDebit[loiDebit]} n'est pas admissible pour les vannes rectangulaires`
                );
            }
            break;
        case StructureType.SeuilRectangulaire:
            if (!(StructureProperties.isCompatibleValues(StructureType.SeuilRectangulaire, loiDebit))) {
                throw new Error(
                    `la loi de débit ${LoiDebit[loiDebit]} n'est pas admissible pour les seuils rectangulaires`
                );
            }
            break;

        default:
            throw new Error(`type de structure ${StructureType[structureType]} non pris en charge`);
    }

    switch (loiDebit) {
        case LoiDebit.Cem88d:
            rectStructPrms.Cd.v = 0.4; // Cd Cem88d qu'on soit en seuil ou vanne
            return new StructureCem88d(rectStructPrms, dbg);

        case LoiDebit.Cem88v:
            rectStructPrms.Cd.v = 0.6; // Cd Cem88d qu'on soit en seuil ou vanne
            return new StructureCem88v(rectStructPrms, dbg);

        case LoiDebit.Cunge80:
            rectStructPrms.Cd.v = 0.6; // Cd Cunge80 qu'on soit en seuil ou vanne
            return new StructureCunge80(rectStructPrms, dbg);

        case LoiDebit.OrificeFree:
            return new StructureOrificeFree(rectStructPrms, dbg);

        case LoiDebit.OrificeSubmerged:
            return new StructureOrificeSubmerged(rectStructPrms, dbg);

        case LoiDebit.WeirFree:
            return new StructureWeirFree(rectStructPrms, dbg);

        case LoiDebit.KIVI:
            const structKiviPrm: StructureKiviParams = new StructureKiviParams(
                8.516, // Q
                101,    // ZDV
                103,    // Z1
                102,    // Z2
                2,      // L
                0.4,  // alpha
                0.001,  // béta
                100);    // ZRAM : cote Radier Amont
            return new StructureKivi(structKiviPrm, dbg);

        default:
            throw new Error(`type de LoiDebit ${LoiDebit[loiDebit]} non pris en charge`);
    }
}
